package webapp

import (
	"database/sql"
	"encoding/json"
	"html/template"
	"log"
	"net"
	"net/http"
	"net/http/httputil"
	"net/url"
	"strconv"
	"strings"
	"slices"
	"time"

	"gontrol/src/logger"
	"gontrol/src/server/api"
)

type App struct {
	Tmpl *template.Template
	DB   *sql.DB
}

func (a *App) renderTemplate(w http.ResponseWriter, name string, data any) {
	t := a.Tmpl.Lookup(name)
	if t == nil {
		log.Printf("Template %s not found", name)
		http.Error(w, "Internal Server Error", http.StatusInternalServerError)
		return
	}

	if err := t.Execute(w, data); err != nil {
		log.Printf("Failed to render %s: %v", name, err)
		http.Error(w, "Internal Server Error", http.StatusInternalServerError)
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Handlers
//
////////////////////////////////////////////////////////////////////////////////

func (a *App) getHomePage(w http.ResponseWriter, r *http.Request) {
	a.renderTemplate(w, "index.html", nil)
}

func (a *App) agentsHandler(w http.ResponseWriter, r *http.Request) {
	parts := strings.Split(strings.TrimPrefix(r.URL.Path, "/agents/"), "/")
	agentId := ""
	if len(parts) > 0 && parts[0] != "" {
		agentId = parts[0]
	}

	switch r.Method {
	case http.MethodDelete:
		api.DeleteAgent(a.DB, w, r, agentId)
		a.listAgents(w,r)
	case http.MethodGet:
		if agentId == "" {
			a.listAgents(w, r)
		} else {
			agent, _ := api.GetAgent(a.DB, w, r, agentId)
			a.renderTemplate(w, "agent_detail.html", agent)
		}
	case http.MethodPost:
		api.CreateAgent(a.DB, w, r)
		a.listAgents(w, r)
	case http.MethodPut:
		api.UpdateAgent(a.DB, w, r, agentId)
		a.listAgents(w, r)
	default:
		http.Error(w, "Method not allowed", http.StatusMethodNotAllowed)
	}
}

func (a *App) listAgents(w http.ResponseWriter, r *http.Request) {
	var agents []api.Agent
	agents, err := api.GetAgents(a.DB)
	if err != nil {
		http.Error(w, "Failed to fetch agents", http.StatusInternalServerError)
		return
	}

	currentAgents := a.getAgentsStatus()

	for i := range agents {

		if  slices.Contains(currentAgents, agents[i].AgentName) {
			agents[i].Status = "Connected"
		} else {
			agents[i].Status = "Disconnected"
		}
	}

	if strings.Contains(r.Header.Get("Accept"), "json") {
		w.Header().Set("Content-Type", "application/json")
		jsonData, err := json.Marshal(agents)
		if err != nil {
			http.Error(w, "Failed to encode agents to JSON", http.StatusInternalServerError)
			return
		}
		w.Write(jsonData)
		return
	}

	a.renderTemplate(w, "agent_list.html", agents)
	// renderTemplate(w, "agent_list.html", agents)
}

func (a *App) logsHandler(w http.ResponseWriter, r *http.Request) {
	// Warning this bit me in the nose: var countStr is []string, but
	// variable = countStr[0] is casted to int automatically
	// when the string is a number. Jesus Christ, this is odd behavior!
	levels := r.URL.Query()["level"]
	countStr := r.URL.Query()["limit"]
	var limit int = 128
	if len(countStr) > 0 {
		parsedCount, err := strconv.Atoi(countStr[0])
		if err == nil {
			limit = parsedCount
		} else {
			http.Error(w, "Invalid count value", http.StatusBadRequest)
			return
		}
	}

	// This enables not only `level` GET parameters but also selecting by paths
	// For example /logs/error is now identical to /logs?level=error
	parts := strings.Split(strings.TrimPrefix(r.URL.Path, "/logs/"), "/")
	if (len(levels) == 0) && len(parts) > 0 && parts[0] != "" {
	 	levels = []string{parts[0]}
	}

	// Call the police... I mean logger
	logs, err := logger.FetchLogs(limit, levels)

	if err != nil {
		http.Error(w, "Error fetching logs", http.StatusInternalServerError)
		return
	}


	a.renderTemplate(w, "logs_partial.html", logs)
}

func (a *App) getAgentsStatus() []string {
	resp, err := http.Get("http://localhost:5555/agentNames")
	if err != nil {
		log.Println("Error fetching agent names:", err)
		logger.InsertLog(logger.Error, "Error fetching agent names from websocketServer")
	}
	defer resp.Body.Close()

	var agentNames []string
	if err := json.NewDecoder(resp.Body).Decode(&agentNames); err != nil {
		log.Println("Error decoding response:", err)
		return []string{}
	}

	return agentNames
}

func (a *App) getAgentNames(w http.ResponseWriter, r *http.Request) {
	api.GetAgentNames(a.DB, w, r)
	return
}

func (a *App) getAgentIds(w http.ResponseWriter, r *http.Request) {
	api.GetAgentIds(a.DB, w, r)
	return
}

// proxyAgentHandler tunnels HTTP and WebSocket traffic to an agent
// selected by ?ip=…&port=… .  It keeps the path *after* /proxyAgent,
// removes the two query parameters, disables HTTP/2 (mandatory for WS),
// and streams with no buffering.
func (a *App) proxyAgentHandler(w http.ResponseWriter, r *http.Request) {
        ip   := r.URL.Query().Get("ip")
        port := r.URL.Query().Get("port")
        if ip == "" || port == "" {
                http.Error(w, "ip and port query parameters are required", http.StatusBadRequest)
                return
        }

        // We leave the scheme "http" even for WebSockets – the Upgrade
        // header does the rest.  (Only cosmetic to change it to "ws”.)
        target, err := url.Parse("http://" + ip + ":" + port)
        if err != nil {
                http.Error(w, "invalid ip/port: "+err.Error(), http.StatusBadRequest)
                return
        }

        proxy := &httputil.ReverseProxy{
                Director: func(req *http.Request) {
                        // Point to the agent
                        req.URL.Scheme = target.Scheme
                        req.URL.Host   = target.Host

                        // Trim the first "/proxyAgent" prefix
                        if strings.HasPrefix(req.URL.Path, "/proxyAgent") {
                                req.URL.Path = strings.TrimPrefix(req.URL.Path, "/proxyAgent")
                                if req.URL.Path == "" {
                                        req.URL.Path = "/"
                                }
                        }

                        // Scrub ip/port from downstream query
                        q := req.URL.Query()
                        q.Del("ip")
                        q.Del("port")
                        req.URL.RawQuery = q.Encode()

                        // Preserve Host (many CLI-style servers care)
                        req.Host = target.Host
                },

                // Critical tweaks for WebSockets
                Transport: &http.Transport{
                        Proxy:               http.ProxyFromEnvironment,
                        DialContext:         (&net.Dialer{Timeout: 30 * time.Second, KeepAlive: 30 * time.Second}).DialContext,
                        ForceAttemptHTTP2:   false,   // MUST be HTTP/1.1 for WS
                        ResponseHeaderTimeout: 0,
                },

                FlushInterval: -1,      // stream bytes immediately
        }

        proxy.ServeHTTP(w, r)
}
